/*
 * image_routines.c Copyright Jan Panteltje 2005 and later.
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "tinyptc.h"

#include "xmars3d.h"


int fill_gamma_table(double gamma)
{
int i;
double da, db, dc;

if(debug_flag)
	{
	fprintf(stderr, "fill_gamma_table(): arg gamma=%.2f\n", gamma);
	}

/* fill in gamma array */
dc = 256.0 / pow(255.0, gamma);
for(i = 0; i < 256; i++)
	{
	/* y = x ** 1/2.2 */

	/* double pow (double BASE, double POWER) */
	da = pow((double)i, gamma); 

	db = da * dc;
	if(db > 255.0) db = 255.0;

	gamma_table[i] = (unsigned char)db;	

	if(debug_flag)
		{
		fprintf(stderr, "i=%d da=%.2f dc=%.2f gamma_table[i]=%u\n", i, da, dc, gamma_table[i]);
		}
	} /* end for fill gamma_table */

return 1;
} /* end function fill_gamma_table */


void set_lcd(int state)
{
int  a;

a = inb(PAR_PORT);

if(state) a |= 4; /* set bit 2 */
else a &= ~4;

outb(a, PAR_PORT);
} /* end function set_lcd */


char *pgm_and_ppm_to_char(FILE *fptr, int *xsize, int *ysize,\
	int start_column, int start_row, int width, int height, int *header_size, int *bytes_per_color)
{
/*
reads pgm file into buffer in BW format.
*/
char *out_ptr;
int a, i;
int maxval;
char *buffer;
char *header_ptr;
size_t line_len;
int header_lines;
long int offset = 0; // -Wall

if(debug_flag)
	{
	fprintf(stderr,\
	"pgm_and_ppm_to_char(): arg fptr=%p start_column=%d start_row=%d width=%d height=%d\n\
header_size=%p *header_size=%d bytes_per_color=%p *bytes_per_color=%d\n",\
	fptr, start_column, start_row, width, height, header_size, *header_size,\
	bytes_per_color, *bytes_per_color); 
	}

if(! fptr) return 0;
if(! xsize) return 0;
if(! ysize) return 0;
if(! header_size) return 0;
if(! bytes_per_color) return 0;

/* read header if we do not have it yet */
if(*header_size == 0)
	{
	/*
	xv writes like this:
	P5
	# CREATOR: XV Version 3.10a  Rev: 12/29/94

	125 107
	255
	*/

	a = fseek(fptr, offset, SEEK_SET);
	if(a != 0)
		{
		fprintf(stderr, "pgm_and_ppm_to_char(): fseek() failed for offset=%lu, aborting.\n", offset);

		exit(1);
		}

	/* parse the header */
	header_lines = 0;
	*bytes_per_color = 0;
	while(1)
		{
		header_ptr = NULL;

//		ssize_t getline (char **LINEPTR, size_t *N, FILE *STREAM)
		a = getline(&header_ptr, &line_len, fptr); 
		if(a == -1) // EOF or error
			{
			fprintf(stderr, "getline() returned %d\n", a);
			break;
			}

		*header_size += a;

		if(header_ptr[0] == '#') continue;

		if(debug_flag)
			{
			fprintf(stderr, "a=%d header_ptr=%s", a, header_ptr);
			}

		if(header_lines == 0)
			{
			/* P5 (pgm) or P6 (ppm)*/
			if(strncmp(header_ptr, "P5", 2) == 0)
				{
				input_image_format = IMAGE_TYPE_PGM;
				*bytes_per_color = 1;
				}
			else if(strncmp(header_ptr, "P6", 2) == 0)
				{
				input_image_format = IMAGE_TYPE_PPM;
				*bytes_per_color = 3;
				}
			else
				{
				fprintf(stderr,\
				"pgm_and_ppm_to_char(): unsupported image_type=%s, aborting.\n", header_ptr);

				exit(1);
				}
			}
		if(header_lines == 1) sscanf(header_ptr, "%d %d", xsize, ysize);
		if(header_lines == 2) sscanf(header_ptr, "%d", &maxval);

		free(header_ptr);

		header_lines++;
		if(header_lines == 3)  break;
		} /* end all header lines */

	if(debug_flag)
		{
		fprintf(stdout,\
		"pgm_and_ppm_to_char(): width=%d height=%d maxval=%d *header_size=%d *bytes_per_color=%d\n",\
		width, height, maxval, *header_size, *bytes_per_color); 
		}
	} /* end if *header_size == 0 */

buffer = (char *) malloc(width * height * *bytes_per_color);
if(! buffer)
	{
	fprintf(stderr, "pgm_and_ppm_to_char(): malloc buffer failed, aborting.\n");

//	exit(1);
	return 0;
	}

out_ptr = buffer;

if(debug_flag)
	{
	fprintf(stderr,\
	"pgm_and_ppm_to_char(): buffer=%p *xsize=%d *ysize=%d width=%d height=%d\n",\
	buffer, *xsize, *ysize, width, height); 
	}

for (i = 0; i < height; i++)
	{
	/* seek to requested row */
	offset = *bytes_per_color * (start_column + ( (start_row + i) * *xsize) ) + *header_size;

//fprintf(stderr, "offset=%d\n", offset);

//	int fseek (FILE *STREAM, long int OFFSET, int WHENCE)
	a = fseek(fptr, offset, SEEK_SET);
	if(a != 0)
		{
//		fprintf(stderr, "pgm_and_ppm_to_char(): fseek() failed for offset=%lu, aborting.\n", offset);

		continue;
//		exit(1);
		}

	/* read width bytes */
	a = fread(out_ptr, sizeof(char), width * *bytes_per_color, fptr);
	if(a < (width * *bytes_per_color) ) break;

	out_ptr += a;

	}/* end for i vertical */

return buffer;
} /* end function pgm_and_ppm_to_char */


int x11_to_pgm(int *buffer, int xpos, int ypos, int width, int height, char *filename)  
{
int b, g, i, j, r, y;
FILE *fptr;
char temp[TEMP_SIZE];
char temp2[4096];
int *iptr;

if(debug_flag)
	{
	fprintf(stderr,\
	"x11_to_pgm(): arg buffer=%p xpos=%d ypos=%d width=%d height=%d filename=%s\n",\
	buffer, xpos, ypos, width, height, filename);
	}

if(! buffer) return 0;
if(! filename) return 0;

/* need more space */
strcpy(temp, filename);

/* append .pgm if not there */
if(strncmp(temp + strlen(filename) - 4, ".pgm", 4) != 0)
	{
	strcat(temp, ".pgm");
	}

/* test of filename exists, ask for overwrite */

fptr = fopen(temp, "r");
if(fptr)
	{
	fclose(fptr);
	sprintf(temp2, "file %s already exists, overwrite (y/n)? ", temp);
	if(debug_flag)
		{
		fprintf(stderr, "%s", temp2);
		}

	text_to_x11(\
	temp2, 70, 160, display_width, display_height, 0xff, 0xff, 0x10, 0x10, 0x10, 0x10, display_data1);

	while(1)
		{
		ptc_update(display_data1);
		set_lcd(0);

		if(key_press == 'y') break;
		if(key_press == 'n')
			{
			key_press = 0;

			return 0;
			}

		usleep(20000);
		}

	key_press = 0;
	} /* end if file exists */

/* open file */
fptr = fopen(temp, "w");
if(! fptr)
	{
	/* report errors */
	sprintf(temp2, "could not open file %s for write", filename);

	if(debug_flag)
		{
		fprintf(stderr, "%s", temp2);
		}

	text_to_x11(\
	temp2, 70, 160, display_width, display_height, 0xff, 0x00, 0x00, 0xff, 0xff, 0xff, display_data1);

	while(1)
		{
		ptc_update(display_data1);
		set_lcd(0);

		if(key_press != 0) return 0;

		usleep(20000);
		}
	}

/* write header */
fprintf(fptr, "P5\n");
fprintf(fptr, "# CREATOR: xmars3d-%s By Jan Panteltje\n", VERSION);
fprintf(fptr, "%d %d\n", width, height);
fprintf(fptr, "255\n");

/* write data */
iptr = buffer;
for(i = 0; i < height; i++)
	{
	for(j = 0; j < width; j++)
		{
		b = *iptr & 0xff;
		g = (*iptr >> 8) & 0xff;
		r = (*iptr >> 16) & 0xff;

		y = (b + g + r) / 3;

		fprintf(fptr, "%c", y);

		iptr++;
		} /* end for j */
	} /* end for i */

/* close file */
fclose(fptr);

return 1;
} /* end function x11_to_pgm */


int x11_to_ppm(int *buffer, int xpos, int ypos, int width, int height, char *filename)  
{
int b, g, i, j, r;
FILE *fptr;
char temp[TEMP_SIZE];
char temp2[4096];
int *iptr;

if(debug_flag)
	{
	fprintf(stderr,\
	"x11_to_ppm(): arg buffer=%p xpos=%d ypos=%d width=%d height=%d filename=%s\n",\
	buffer, xpos, ypos, width, height, filename);
	}

if(! buffer) return 0;
if(! filename) return 0;

/* need more space */
strcpy(temp, filename);

/* append .ppm if not there */
if(strncmp(temp + strlen(filename) - 4, ".ppm", 4) != 0)
	{
	strcat(temp, ".ppm");
	}

/* test of filename exists, ask for overwrite */

fptr = fopen(temp, "r");
if(fptr)
	{
	fclose(fptr);
	sprintf(temp2, "file %s already exists, overwrite (y/n)? ", temp);
	if(debug_flag)
		{
		fprintf(stderr, "%s", temp2);
		}

	text_to_x11(\
	temp2, 70, 160, display_width, display_height, 0xff, 0xff, 0x10, 0x10, 0x10, 0x10, display_data1);

	while(1)
		{
		ptc_update(display_data1);
		set_lcd(0);

		if(key_press == 'y') break;
		if(key_press == 'n')
			{
			key_press = 0;

			return 0;
			}

		usleep(20000);
		}

	key_press = 0;
	} /* end if file exists */

/* open file */
fptr = fopen(temp, "w");
if(! fptr)
	{
	/* report errors */
	sprintf(temp2, "could not open file %s for write", filename);

	if(debug_flag)
		{
		fprintf(stderr, "%s", temp2);
		}

	text_to_x11(\
	temp2, 70, 160, display_width, display_height, 0xff, 0x00, 0x00, 0xff, 0xff, 0xff, display_data1);

	while(1)
		{
		ptc_update(display_data1);
		set_lcd(0);

		if(key_press != 0) return 0;

		usleep(20000);
		}
	}

/* write header */
fprintf(fptr, "P6\n");
fprintf(fptr, "# CREATOR: xmars3d-%s By Jan Panteltje\n", VERSION);
fprintf(fptr, "%d %d\n", width, height);
fprintf(fptr, "255\n");

/* write data */
iptr = buffer;
for(i = 0; i < height; i++)
	{
	for(j = 0; j < width; j++)
		{
		b = *iptr & 0xff;
		g = (*iptr >> 8) & 0xff;
		r = (*iptr >> 16) & 0xff;

		fprintf(fptr, "%c%c%c", r, g, b);

		iptr++;
		} /* end for j */
	} /* end for i */

/* close file */
fclose(fptr);

return 1;
} /* end function x11_to_ppm */



